
#include "poppet.h"


#include <io.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

enum cmd {VAR_BOOL, VAR_INT, VAR_STR};

typedef struct {
  char name[SHORT_STRING];
  int type;
  void *data;
} Command;



Command variable[] =
{
  {"beep_new", VAR_BOOL, &BeepNew},
  {"pop_delete", VAR_BOOL, &PopDelete},
  {"pop_last", VAR_BOOL, &PopLast},
  {"pop_port", VAR_INT, &PopPort},
  {"pop_checkinterval", VAR_INT, &PopInterval},
  {"tmpdir", VAR_STR, &TmpDir},
  {"folder", VAR_STR, &Folder},
  {"mbox_type", VAR_STR, &MboxType},
  {"pop_host", VAR_STR, &PopHost},
  {"pop_user", VAR_STR, &PopUser},
  {"pop_pass", VAR_STR, &PopPass},
  {"alternates", VAR_STR, &Alternates},
  {"default_hook", VAR_STR, &DefaultHook},
  {"", 0, 0}
};


char *get_next_word(char *s, int eq, char *cmd, int len)
{
  int i = 0;
  int quoted = 0;

  while (s && *s && i<len-1 &&
	 (quoted!=0 || 
	  (quoted==0 && (!isspace(*s) && *s != '#' && *s != '%' && *s != ';' && (eq || *s != '='))))) {
    if (*s == '"' && (quoted == 0 || quoted == 1)) 
      quoted = quoted ^ 1;
    else if (*s == '\'' && (quoted == 0 || quoted == 2)) 
      quoted = quoted ^ 2;
    else
      cmd[i++] = *s;
    s++;
  }
  cmd[i] = 0;
  return s;
}

char *skip_blanks(char *s, int eq)
{
  while (s && *s && (isspace(*s) || (eq && *s=='='))) {
    s++;
  }
  return s;
}


int parse_line(char *s) 
{
  char word[3][SHORT_STRING];
  int w,i,j;
  
  w = i = j = 0;

  for (w=0; s && *s && w<3; w++) {
    s = skip_blanks(s, w==1);
    s = get_next_word(s, w!=0, word[w], SHORT_STRING);
    word[w][SHORT_STRING-1] = 0;
    if (w == 0 && strcmp(word[0], "set") == 0)
      w--;
  }

  if (word[0][0] == 0)
    return 0;

  if (strcmp(word[0], "filter-hook") == 0)
  {
    int ok;
    int cflags = REG_NEWLINE + REG_ICASE + REG_NOSUB;
    Filter *ptr;

    log_printf("  filter-hook \"%s\" \"%s\"\n", word[1], word[2]);

    ptr = (Filter*) calloc( 1, sizeof(Filter) );
    if (!ptr) {
      log_printf("Memory allocation error!\n");
      return 0;
    }
    ok = regcomp (&(ptr->preg), word[1], cflags);
    if (ok != 0) {
      char buf[SHORT_STRING];
      regerror(ok, &(ptr->preg), buf, sizeof(buf));
      log_printf("regerror(%d) = '%s'\n", ok, buf);
      return -1;
    }
    strncpy( ptr->mbox, word[2], SHORT_STRING );
    ptr->num = 0;

    if (FilterRoot == NULL)
      FilterRoot = ptr;
    else 
    {
      Filter *tmp = FilterRoot;
      while (tmp && tmp->next)
	tmp = tmp->next;
      if (tmp)
	tmp->next = ptr;
      else
	FilterRoot = ptr;
    }
  }
  else 
  {
    for (i=0; variable[i].name[0]; i++) {
      if (strcmp(variable[i].name, word[0]) == 0)
	break;
    }
    if (variable[i].name[0] && w>1 && word[1][0])
    {
      switch (variable[i].type) {
      case VAR_BOOL:
	if (strcmp(word[1], "yes") == 0)
	  *((int*)variable[i].data) = 1;
	else if (strcmp(word[1], "no") == 0)
	  *((int*)variable[i].data) = 0;
	else
	{
	  log_printf("Bad boolean value '%s'\n", word[1]);
	  return -1;
	}
	log_printf("  set %s = %s\n", 
		   variable[i].name, 
		   *((int*)variable[i].data) ? "yes" : "no");
	break;

      case VAR_INT:
	{ int j;
	for (j=0; word[1][j]; j++)
	  if (!isdigit(word[1][j])) {
	    log_printf("Bad integer value '%s'\n", word[1]);
	    return -1;
	  }
	}
	*(int*)variable[i].data = atoi(word[1]);
	log_printf("  set %s = %d\n", 
		   variable[i].name, 
		   *((int*)variable[i].data));
	break;

      case VAR_STR:
	strncpy( (char*)variable[i].data, word[1], SHORT_STRING );
	if (strcmp(variable[i].name, "mbox_type")==0) {
	  if (strcmp((char*)variable[i].data, "mbox") &&
	      strcmp((char*)variable[i].data, "maildir")) {
	    log_printf("Bad mbox_type '%s'\n", (char*)variable[i].data);
	    strcpy((char*)variable[i].data, "");
	    return -1;
	  }
	}
	log_printf("  set %s = \"%s\"\n", 
		   variable[i].name, 
		   variable[i].data == &PopPass ? "<hidden>" :
		   (char*)variable[i].data);
	break;
      }

    }
    else
    {
      if (variable[i].name[0])
	log_printf("Missing value\n");
      else
	log_printf("Unknown variable '%s'\n", word[0]);
      return -1;
    }
  }

  return 0;
}

int read_poppetrc (char *Poppetrc)
{
  char buf[4096];
  FILE *fp;
  char *ch;
  int line = 0;
  int errors = 0;

  log_printf("\nOpening Poppetrc.\n");

  fp = fopen(Poppetrc, "r");
  if (fp == NULL) {
    log_printf ("Unable to open Poppetrc, \"%s\"\n", Poppetrc);
    return -1;
  }

  log_printf("Reading Poppetrc.\n");

  ch = fgets(buf, sizeof(buf), fp);
  while (ch != NULL) {
    line++;
    if (parse_line (buf)) {
      log_printf ("Syntax error (line %d): %s", line, buf);
      errors++;
    }
    ch = fgets(buf, sizeof(buf), fp);
  }

  log_printf("Closing Poppetrc.\n\n");
  fclose(fp);

  if (Folder[0] == '~') {
    snprintf(buf, sizeof(buf), "%s%s", Homedir, Folder+1);
    strncpy(Folder, buf, sizeof(Folder));
  }

  return errors ? -1 : 0;
}


int init_poppet(void)
{
  char Poppetrc[_POSIX_PATH_MAX] = "";
  char *p;

  PopPort = 110;
  strcpy(Folder, "~/Mail");
  FilterRoot = NULL;

  p = getenv ("COMPUTERNAME");
  if (!p) p = getenv ("HOSTNAME");
  if (!p) p = getenv ("USERDOMAIN");

  if (p)
    strncpy(Hostname, p, sizeof(Hostname));
  else
    strncpy(Hostname, "LOCALHOST", sizeof(Hostname));

  log_printf("  %%HOSTNAME%% = %s\n", Hostname);

  if (GetModuleFileName( NULL, PoppetRoot, sizeof(PoppetRoot)) == 0)
    ;
  /* strncpy (PoppetRoot, argv[0], _POSIX_PATH_MAX); */
  
  if ((p = strrchr (PoppetRoot, '\\')))
    *p = '\0';

  if (!PoppetRoot[0] || getenv ("POPPET_ROOT"))
    strncpy(PoppetRoot, (p = getenv ("POPPET_ROOT")) ? p : "\\poppet", _POSIX_PATH_MAX);

  log_printf("  %%POPPET_ROOT%% = %s\n", PoppetRoot);

  p = getenv ("HOME");
  if (!p) p = getenv ("USERPROFILE");
  if (!p) p = getenv ("HOMEPATH");
  strncpy(Homedir, p, sizeof(Homedir));

  log_printf("  %%HOME%% = %s\n", Homedir);

  snprintf (Poppetrc, sizeof(Poppetrc), "%s\\.poppetrc", Homedir);
  if (access(Poppetrc, F_OK) == -1)
    snprintf (Poppetrc, sizeof(Poppetrc), "%s\\poppet.rc", Homedir);
  if (access(Poppetrc, F_OK) == -1)
    snprintf (Poppetrc, sizeof(Poppetrc), "%s\\.poppetrc", PoppetRoot);
  if (access(Poppetrc, F_OK) == -1)
    snprintf (Poppetrc, sizeof(Poppetrc), "%s\\poppet.rc", PoppetRoot);
  if (access(Poppetrc, F_OK) == -1)
    Poppetrc[0] = 0;

  if (!Poppetrc[0] || getenv ("POPPET_RC"))
    strncpy(Poppetrc, (p = getenv ("POPPET_RC")) ? p : "poppet.rc", _POSIX_PATH_MAX);

  log_printf("  %%POPPET_RC%% = %s\n", Poppetrc);

  if (!Poppetrc[0]) {
    log_printf("Missing Poppetrc.\n");
    return -1;
  }

  FilterError = (Filter*) calloc( 1, sizeof(Filter) );
  if (!FilterError) {
    log_printf("Memory allocation error!\n");
    return -1;
  }
  snprintf(FilterError->mbox, sizeof(FilterError->mbox), "<ERROR>");
  FilterRoot = FilterError;

  if (read_poppetrc (Poppetrc) < 0) {
    log_printf("Errors in Poppetrc.\n");
    return -1;
  }

  return 0;
}
