
#include "poppet.h"
#include "resources.h"

#include <sys/types.h>
#include "_regex.h"

#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <windows.h>
#include <winsock.h>
#include <wchar.h>
#include <sys/utime.h>
#include <sys/stat.h>
#include <errno.h>

#include <stdarg.h>
# define HAVE_STDARGS    /* let's hope that works everywhere (mj) */
# define VA_LOCAL_DECL   va_list ap
# define VA_START(f)     va_start(ap, f)
# define VA_SHIFT(v,t)  ;   /* no-op for ANSI */
# define VA_END          va_end(ap)


#ifndef NDEBUG

int SEND(SOCKET s, char *buf, int len) {
  /* log_printf("send: '%s'\n", buf); */
  return send(s, buf, len, 0);
}

int RECV(SOCKET s, char *buf, int len) {
  int tmp = recv(s, buf, len, 0);
  /* log_printf("recv: '%s'\n", buf); */
  return tmp;
}


#ifdef CONSOLE
DebugLog log_printf = printf;
#else
char *log_text = NULL;
int log_size = 1;

static int print_log(const char *str, ...) 
{ 
  int i, j, len;
  char tmpline[HUGE_STRING];
  VA_LOCAL_DECL;
    
  VA_START (str);
  VA_SHIFT (str, char *);
  VA_SHIFT (count, size_t );
  VA_SHIFT (str, char *);
  len = vsnprintf(tmpline, sizeof(tmpline), str, ap);
  VA_END;
  
  for (i=j=0; i<=len; i++)
    if (tmpline[i]=='\n')
      j++;

  if (j>0)
    for (i=len; i>=0; i--) {
      tmpline[i+j] = tmpline[i];
      if (tmpline[i] == '\n') {
	j--;
	len++;
	tmpline[i+j] = '\r';
      }
    }

  if (len>0) {
    char *tmpbuf;
    tmpbuf = malloc(log_size + len);
    if (tmpbuf) {
      log_size += len;
      snprintf(tmpbuf, log_size, "%s%s", log_text ? log_text:"", tmpline);
      if (log_text)
	free(log_text);
      log_text = tmpbuf;
    }
  }
  return 0; 
}
DebugLog log_printf = print_log;
#endif
#else
static int noprint(const char *str, ...) { return 0; }
DebugLog log_printf = noprint;
#endif


char *cleanStr(char *buf)
{
  if (buf && *buf) {
    char *p, *q;
    int i, len;

    p = q = buf;
    len = strlen(buf);

    for (i=0; *p && i<len;) {
      if (*p == '\r') { p++; i++; }
      if (*p == '\n' && isspace(*(p+1))) { p++; i++; }
      if (++i<len)
	*(q++) = *(p++);
    }
    if (q > buf && *(q-1)=='\n')
      q--;
    *q = '\0';
  }

  return buf;
}


void file_mktemp(char *s, size_t l)
{
  snprintf (s, l, "%s/popXXXXXX", TmpDir);
  mktemp (s);
}


int file_cat(char *dest, char *src, int fsize, int lines)
{
  struct stat st = {0};
  char buffer[4096];
  FILE *fp_in;
  FILE *fp_out;
  int i;

  if ( (fp_in = fopen(src, "rb")) == NULL)
    return -1;

  stat (dest, &st);

  if ( (fp_out = fopen(dest, "ab+")) == NULL) {
    fclose(fp_in);
    return -1;
  }

  if (lines > 0 && fsize > 0) {
    char *ch;
    ch = fgets( buffer, sizeof(buffer), fp_in );
    while (ch!=NULL && strcmp(buffer, "\n") && strcmp(buffer, "\r\n")) {
      fputs( buffer, fp_out );
      ch = fgets( buffer, sizeof(buffer), fp_in );
    }
    fprintf( fp_out, "Content-Length: %d\n", fsize);
    fprintf( fp_out, "Lines: %d\n\n", lines);
  }

  i = fread(buffer, 1, sizeof(buffer), fp_in);
  while (i > 0) {
    fwrite(buffer, 1, i, fp_out);
    i = fread(buffer, 1, sizeof(buffer), fp_in);
  }

  fclose(fp_in);
  fclose(fp_out);

  if (st.st_atime)
  {
    struct utimbuf utim;

    utim.actime = st.st_atime;
    utim.modtime = time (NULL);
    utime (dest, &utim);
  }

  return i;
}

static int 
move_message(char *tmpfname, int fsize, int lines, char *dest)
{
  enum mbox_type {mbox, maildir} type;
  struct stat sb;
  int ret, len;

  if (strcmp(MboxType, "maildir"))
    type = maildir;
  else 
    type = mbox;

  len = strlen(dest);
  if (len > 0 && (dest[len-1]=='\\' || dest[len-1]=='/'))
    type = maildir;
  else 
    if (stat (dest, &sb)) {
      if (errno != ENOENT) {
	log_printf ("ERROR: stat %s: %s (errno %d)\n", dest,
		    strerror (errno), errno);
	return -1;
      }
    }
    else {
      if (S_ISDIR (sb.st_mode))
	type = maildir;
      else 
	type = mbox;
    }

  if (type == maildir)
    ret = maildir_save( dest, tmpfname, fsize, lines );
  else
    ret = file_cat( dest, tmpfname, fsize, lines );

  return ret;
}

void size2str( int size, char *msg)
{
    if (size < 10*1024) {
        int tmp = (10*size) / 1024;
	msg[0] = '0' + tmp / 10;
	msg[1] = '.';
	msg[2] = '0' + tmp % 10;
	msg[3] = 'k';
	msg[4] = 0;
    }
    else if (size < 100*1024) {
        snprintf(msg, 5, "%dk", size / 1024);
    }
    else if (size < 10*1024*1024) {
        int tmp = (10*size) / (1024*1024);
	msg[0] = '0' + tmp / 10;
	msg[1] = '.';
	msg[2] = '0' + tmp % 10;
	msg[3] = 'M';
	msg[4] = 0;
    }
    else {
        snprintf(msg, 5, "%dM", size / (1024*1024));	
    }
}

static int pop_get_line (int fd, char *s, int len)
{
  char ch; /* , *ptr=s; */
  int bytes = 0;

  while (RECV (fd, &ch, 1) > 0)
  {
    *s++ = ch;
    bytes++;
    if (ch == '\n')
    {
      *s = 0;
      /* log_printf("recv! '%s'\n", ptr); */
      return (bytes);
    }
    /* make sure not to overwrite the buffer */
    if (bytes == len - 1)
    {
      *s = 0;
      /* log_printf("recv? '%s'\n", ptr); */
      return bytes;
    }
  }
  *s = 0;
  /* log_printf("!recv '%s'\n", ptr); */
  return (-1);
}


char TimeBuf [SHORT_STRING];


int pop_fetch_mail (char *errstr, int errlen)
{
  int ANIM_STEP=10;
  struct sockaddr_in sin;
#if SIZEOF_LONG == 4
  long n;
#else
  int n;
#endif
  struct hostent *he;
  char buffer[HUGE_STRING];
  char msgbuf[STRING];
  int s, i, j, last = 0, msgs, bytes, err = 0;
  int ret = -1;
  time_t t;
  struct tm *tm;
  char tmpfname [SHORT_STRING];
  char fname[_POSIX_PATH_MAX];
  FILE *fp;
  char *p, *q, *r;
  int ok, size;
  Filter *match;
  long fsize;
  int icon = -1;
  int lines = -1;

  snprintf(errstr, errlen, "<Poppet>");

  if (!PopHost || !*PopHost)
  {
    snprintf(errstr, errlen, "POP host is not defined.");
    return -1;
  }

  if (!PopUser || !*PopUser)
  {
    snprintf(errstr, errlen, "No POP username is defined.");
    return -1;
  }

  if (!PopPass || !*PopPass)
  {
    snprintf(errstr, errlen, "No POP password is defined.");
    return -1;
  }

  s = socket (AF_INET, SOCK_STREAM, IPPROTO_IP);

  memset ((char *) &sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons (PopPort);

  if ((n = inet_addr (PopHost)) == -1)
  {
    /* Must be a DNS name */
    if ((he = gethostbyname (PopHost)) == NULL)
    {
      snprintf(errstr, errlen, "Could not find address for host %s.", PopHost);
      return -1;
    }
    memcpy ((void *)&sin.sin_addr, *(he->h_addr_list), he->h_length);
  }
  else
    memcpy ((void *)&sin.sin_addr, (void *)&n, sizeof(n));
  
  t = time(NULL);
  tm = localtime(&t);
  strftime(TimeBuf, sizeof(TimeBuf), "%H:%M:%S", tm);
  log_printf("Connecting to %s [%s]\n", inet_ntoa (sin.sin_addr), TimeBuf);

  errno = 0;
  if (connect (s, (struct sockaddr *) &sin, sizeof (struct sockaddr_in)) == -1)
  {
    log_printf("Unable to connect to %s\n", inet_ntoa (sin.sin_addr));
    perror ("connect");
    snprintf(errstr, errlen, "Unable to connect to %s", inet_ntoa (sin.sin_addr));
    return -1;
  }
  
  if (pop_get_line (s, buffer, sizeof (buffer)) == -1)
    goto fail;

  if (strncmp (buffer, "+OK", 3) != 0)
  {
    cleanStr(buffer);
    log_printf("%s\n", buffer);
    snprintf(errstr, errlen, "%s", buffer);
    goto finish;
  }

  snprintf (buffer, sizeof(buffer), "user %s\r\n", PopUser);
  SEND (s, buffer, strlen (buffer));

  if (pop_get_line (s, buffer, sizeof (buffer)) == -1)
    goto fail;

  if (strncmp (buffer, "+OK", 3) != 0)
  {
    cleanStr(buffer);
    log_printf("%s\n", buffer);
    snprintf(errstr, errlen, "%s", buffer);
    goto finish;
  }
  
  snprintf (buffer, sizeof(buffer), "pass %s\r\n", PopPass);
  SEND (s, buffer, strlen (buffer));
  
  if (pop_get_line (s, buffer, sizeof (buffer)) == -1)
    goto fail;

  if (strncmp (buffer, "+OK", 3) != 0)
  {
    cleanStr(buffer);
    log_printf("%s%s\n", buffer[0] ? "< " : "", buffer[0] ? buffer : "Server closed connection!");
    snprintf(errstr, errlen, "%s", buffer[0] ? buffer : "Server closed connection!");
    goto finish;
  }
  
  /* find out how many messages are in the mailbox. */
  SEND (s, "stat\r\n", 6);
  
  if (pop_get_line (s, buffer, sizeof (buffer)) == -1)
    goto fail;

  if (strncmp (buffer, "+OK", 3) != 0)
  {
    cleanStr(buffer);
    log_printf("%s\n", buffer);
    snprintf(errstr, errlen, "%s", buffer);
    goto finish;
  }
  
  sscanf (buffer, "+OK %d %d", &msgs, &bytes);
  ret = 0;

  if (msgs == 0)
  {
    log_printf("No new mail in POP mailbox.\n");
    snprintf(errstr, errlen, "No new mail in POP mailbox.");
    goto finish;
  }


  /* only get unread messages */
  if (PopLast)
  {
    SEND (s, "last\r\n", 6);
    if (pop_get_line (s, buffer, sizeof (buffer)) == -1)
      goto fail;
    
    if (strncmp (buffer, "+OK", 3) == 0)
      sscanf (buffer, "+OK %d", &last);
    else
      /* ignore an error here and assume all messages are new */
      last = 0;
  }
  
  if (msgs == last)
  {
    log_printf("No new mail in POP mailbox.\n");
    snprintf(errstr, errlen, "No new mail in POP mailbox.");
    goto finish;
  }

  snprintf (msgbuf, sizeof (msgbuf),
	    "Reading %d new of %d %s (%d bytes). ",
	    msgs-last, msgs, msgs == 1 ? "message" : "messages", bytes);
  log_printf("%s\n", msgbuf);

  for (i = last + 1 ; i <= msgs ; i++)
  {
    log_printf(" ----\n");
    fname[0] = 0;
    snprintf (buffer, sizeof(buffer), "top %d 0\r\n", i);
    SEND (s, buffer, strlen (buffer));

    if (pop_get_line (s, buffer, sizeof (buffer)) == -1)
    {
      goto fail;
    }
    
    if (strncmp (buffer, "+OK", 3) == 0)
      sscanf (buffer, "+OK %d", &size);
    else {
      cleanStr(buffer);
      log_printf("%s\n", buffer);
      snprintf(errstr, errlen, "%s", buffer);
      continue;
    }

    {
      int a, b;
      size = size + 5;			/*  '\r\n.\r\n'  */
      p = q = r = malloc(size+2);
      *(p++) = '\n';
      a = 0;
      do {
	if ( (b = RECV (s, p, size - a)) > 0 ) {
	  a = a + b;
	  p = p + b;
	}
	*p = 0;
	q = strstr(r, "\r\n.\r\n");
      } while (b>0 && a < size && !q);
      p = q = r;

      /*log_printf("<<<%s>>> %d %d %d\n", r, size, a, b);*/
    }

    size = size - 5;
    for (j=0; *p && j<size;) {
      if (*p == '\r') { p++; j++; }
      if (*p == '\n' && isspace(*(p+1))) { p++; j++; }
      if (++j<size)
	*(q++) = *(p++);
    }
    *q = '\0';

    /*log_printf("[[[%s]]]\n", r);*/

    match = FilterRoot;
    while (match) {
      if (match!=FilterError)
	ok = regexec (&(match->preg), r, 0, NULL, REG_NOTBOL | REG_NOTEOL);
      else 
	ok = -1;
      if (ok == 0) {
	match->num++;
	break;
      }
      else 
	match = match->next;
    }

    p = strstr(r, "\nFrom:");
    if (p) {
      q = p + 1;
      while (*q && *q!='\n')
	q++;
      *q = 0;
      log_printf("  %s\n", p+1);
      *q = '\n';
    }

    p = strstr(r, "\nSubject:");
    if (p) {
      q = p + 1;
      while (*q && *q!='\n')
	q++;
      *q = 0;
      log_printf("  %s\n", p+1);
      *q = '\n';
    }

    {
      char msg[5];
      size2str(size, msg);
      log_printf("  Size: %s\n", msg);
    }

    free(r);

    if (match && strcmp(match->mbox, "<delete>") == 0)
      goto delete_mail;
    
    if (!match || strcmp(match->mbox, "<keep>") == 0) {
      log_printf("  [Kept]\n");
      continue;
    }
      
    snprintf (buffer, sizeof(buffer), "retr %d\r\n", i);
    SEND (s, buffer, strlen (buffer));

    if (pop_get_line (s, buffer, sizeof (buffer)) == -1)
    {
      goto fail;
    }

    if (strncmp (buffer, "+OK", 3) != 0)
    {
      cleanStr(buffer);
      log_printf("%s\n", buffer);
      snprintf(errstr, errlen, "%s", buffer);
      break;
    }

    file_mktemp(tmpfname, sizeof(tmpfname));
    fp = fopen(tmpfname, "wb+");
    if (!fp) {
      log_printf("Unable to create temporary file '%s'!\n", tmpfname);
      snprintf(errstr, errlen, "Unable to create temporary file '%s'!", tmpfname);
      return -1;
    }

    t = time(NULL);
    fprintf (fp, "From %s %s", PopUser, ctime (&t));

    fsize = 0;
    lines = -1;
    icon = -1;
    ANIM_STEP = 1;
    /* Now read the actual message. */
    for(;;)
    {
      char *p;
      int chunk;
      char msg[STRING];

      icon = (icon + 1) % (16*ANIM_STEP);
      if (icon % ANIM_STEP == 0) {
	if (fsize < 10*1024)
	{
	  int tmp = (10*fsize) / 1024;
	  msg[0] = '0' + tmp / 10;
	  msg[1] = '.';
	  msg[2] = '0' + tmp % 10;
	  msg[3] = 'k';
	  msg[4] = 0;
	}
	else if (fsize < 100*1024) 
	{
	  snprintf(msg, sizeof(msg), "%dk", fsize / 1024);
	}
	else if (fsize < 10*1024*1024) {
	  int tmp = (10*fsize) / (1024*1024);
	  msg[0] = '0' + tmp / 10;
	  msg[1] = '.';
	  msg[2] = '0' + tmp % 10;
	  msg[3] = 'M';
	  msg[4] = 0;
	}
	else
	{
	  snprintf(msg, sizeof(msg), "%dM", fsize / (1024*1024));	
	}
	snprintf(msgbuf, sizeof(msgbuf), "Downloading %d of %d Messages [%s]", 
		 i-last, msgs-last, msg );
	InstallTrayIcon( NIM_MODIFY, TRAYICON_ENV_01+icon/ANIM_STEP, msgbuf );
	if (icon==0)
	  ANIM_STEP++;
      }
      if ((chunk = pop_get_line (s, buffer, sizeof (buffer))) == -1)
      {
	log_printf("Error reading message!\n");
	snprintf(errstr, errlen, "Error reading message!");
	err = 1;
	break;
      }

      /* check to see if we got a full line */
      if (buffer[chunk-2] == '\r' && buffer[chunk-1] == '\n')
      {
	if (strcmp(".\r\n", buffer) == 0)
	{
	  /* end of message */
	  break;
	}
	if (lines > -1) {
	  lines++;
	}
	else
	{
	  if (strcmp("\r\n", buffer) == 0)
	    lines = 0;
	}

	/* change CRLF to just LF */
	buffer[chunk-2] = '\n';
	buffer[chunk-1] = 0;
	chunk--;

	/* see if the line was byte-stuffed */
	if (buffer[0] == '.')
	{
	  p = buffer + 1;
	  chunk--;
	}
	else
	  p = buffer;
	if (lines > 0)
	  fsize += chunk;
      }
      else {
	if (lines > 0)
	  fsize += chunk;
	p = buffer;
      }
      
      fwrite (p, 1, chunk, fp);
    }
    if (!err)
      ret++;

    fprintf (fp, "\n");
    fclose(fp);

    snprintf(msgbuf, sizeof(msgbuf), "Downloaded %d of %d Messages", 
	     i-last, msgs-last);
    InstallTrayIcon( NIM_MODIFY, TRAYICON_ENV_09, msgbuf );

    if (err)
      break;

    if (match->mbox[0] == '~')
      snprintf(fname, sizeof(fname), "%s\\%s", Homedir, match->mbox+1);
    else if (match->mbox[0] == '=' || match->mbox[0] == '+')
      snprintf(fname, sizeof(fname), "%s\\%s", Folder, match->mbox+1);
    else
      snprintf(fname, sizeof(fname), "%s", match->mbox);

    if (move_message( tmpfname, fsize, lines, fname) == 0)
      remove(tmpfname);
    else {
      ret--;
      FilterError->num++;
      log_printf("Unable to move message from '%s' to '%s'\n", tmpfname, fname);
      snprintf(errstr, errlen, "Unable to move message from %s to %s!", tmpfname, fname);
      continue;
    }

    if (PopDelete)
    {
  delete_mail:
    snprintf(msgbuf, sizeof(msgbuf), "Downloaded %d of %d Messages", 
	     i-last, msgs-last);
    InstallTrayIcon( NIM_MODIFY, TRAYICON_ENV_09, msgbuf );

      /* delete the message on the server */
      snprintf (buffer, sizeof(buffer), "dele %d\r\n", i);
      SEND (s, buffer, strlen (buffer));

      /* eat the server response */
      pop_get_line (s, buffer, sizeof (buffer));
      if (strncmp (buffer, "+OK", 3) != 0)
      {
	err = 1;
	cleanStr(buffer);
	log_printf("%s\n", buffer);
	snprintf(errstr, errlen, "%s", buffer);
	break;
      }
      if (*fname)
	log_printf("  [Moved] -> %s\n", fname);
      else
	log_printf("  [Deleted]\n");
    }
    else
      log_printf("  [Copied] -> %s\n", fname);
  }
  log_printf(" ====\n");

  if (err)
  {
    /* make sure no messages get deleted */
    SEND (s, "rset\r\n", 6);
    pop_get_line (s, buffer, sizeof (buffer)); /* snarf the response */
  }

finish:

  /* exit gracefully */
  SEND (s, "quit\r\n", 6);
  pop_get_line (s, buffer, sizeof (buffer)); /* snarf the response */
  CLOSESOCKET (s);
  return ret;

  /* not reached */

fail:

  log_printf("Server closed connection!\n");
  snprintf(errstr, errlen, "Server closed connection!");
  CLOSESOCKET (s);
  return -1;
}

