#include "poppet.h"
#include "resources.h"
#include <windows.h>
#include <time.h>

#define WM_TRAYNOTIFY   (WM_USER + 1)
#define IDC_TRAYICON    1000

static HWND myHwnd;
static int mails = 0;
static char *history = NULL;

static int busyDeliver = 0;
static int busyHistory = 0;
static int busyAbout = 0;

int quick_exit = 0;
extern char TimeBuf[];

BOOL CALLBACK
AboutDlgProc( HWND hwnd,
	      UINT Message,
	      WPARAM wParam,
	      LPARAM lParam )
{
    switch (Message) {
      case WM_INITDIALOG:
        return TRUE;

      case WM_COMMAND:
        switch (LOWORD(wParam)) {
	  case IDOK:
	    EndDialog( hwnd, IDOK );
	    break;
	  case IDCANCEL:
	    EndDialog( hwnd, IDCANCEL );
	    break;
	}
	break;
	
      default:
        return FALSE;
    }
    return TRUE;
}


void buildString( char *msg, int msglen )
{
  char buf[LONG_STRING];
  Filter *tmp = FilterRoot;

  *msg = 0;

  while (tmp) {
    if (tmp->num) {
      snprintf( buf, sizeof(buf), "%s%s\t%3d", *msg ? "\r\n":"", tmp->mbox, tmp->num);
      strncat( msg, buf, msglen );
      tmp->num = 0;
    }
    tmp = tmp->next;
  }
}


BOOL CALLBACK
StatusDlgProc( HWND hwnd,
	       UINT Message,
	       WPARAM wParam,
	       LPARAM lParam )
{
    switch (Message) {
      case WM_INITDIALOG:
	{
	  int i = 0, update, len;
	  char *buf, *tmpbuf;
	  Filter *tmp = FilterRoot;

	  while (tmp) {
	    i++;
	    tmp = tmp->next;
	  }
	  i = i*(LONG_STRING);

	  buf = malloc( i );
	  if (buf) {
	    time_t t;
	    struct tm *tm;
	    char tmptime [SHORT_STRING];

	    t = time(NULL);
	    tm = localtime(&t);
	    strftime(tmptime, sizeof(tmptime), "%H:%M:%S", tm);

	    buildString( buf, i );
	    if (*buf == 0) {
	      snprintf(buf, i, "No new mail in POP mailbox.");
	      update = 0;
	    }
	    else 
	      update = 1;

	    len = 30 + (history ? strlen(history): 0) + strlen(buf);
	    tmpbuf = malloc(len);

	    snprintf(tmpbuf, len, "[%s]\r\n%s%s%s", 
		     update ? TimeBuf : tmptime,
		     buf, 
		     history ? "\r\n====\r\n" : "",
		     history ? history : "");

	    if (update) {
	      free(history);
	      history = tmpbuf;
	    }

	    SetDlgItemText(hwnd, IDC_MESSAGES, tmpbuf);

	    free(buf);
	    if (!update)
	      free(tmpbuf);
	  }
	  mails = 0;
	  InstallTrayIcon(NIM_MODIFY, TRAYICON_ENV_01, "Poppet");
	  return TRUE;
	}

      case WM_COMMAND:
        switch (LOWORD(wParam)) {
	  case IDOK:
	    EndDialog( hwnd, IDOK );
	    break;
	  case IDCANCEL:
	    EndDialog( hwnd, IDCANCEL );
	    break;
	}
	break;
	
      default:
        return FALSE;
    }
    return TRUE;
}


#if !defined(NDEBUG) && !defined(CONSOLE)
char *log_text;

BOOL CALLBACK
LogDlgProc( HWND hwnd,
	    UINT Message,
	    WPARAM wParam,
	    LPARAM lParam )
{
    switch (Message) {
      case WM_INITDIALOG:
	{
	  SetDlgItemText(hwnd, IDC_MESSAGES, log_text);
	  return TRUE;
	}

      case WM_COMMAND:
        switch (LOWORD(wParam)) {
	  case IDOK:
	    EndDialog( hwnd, IDOK );
	    break;
	  case IDCANCEL:
	    EndDialog( hwnd, IDCANCEL );
	    break;
	}
	break;
	
      default:
        return FALSE;
    }
    return TRUE;
}
#endif

void InstallTrayIcon( DWORD dwMsgType, UINT iconID, char *strToolTip )
{
  HICON hIcon;
  NOTIFYICONDATA tnd;

        hIcon =  
	  (HICON)LoadImage(thehInstance,
			   MAKEINTRESOURCE(iconID), 
			   IMAGE_ICON, 16, 16, 0);
        tnd.cbSize              = sizeof(NOTIFYICONDATA);
	tnd.hWnd                = myHwnd;
	tnd.uID                 = IDC_TRAYICON;
	tnd.uFlags              = NIF_MESSAGE | NIF_ICON | NIF_TIP;
	tnd.uCallbackMessage    = WM_TRAYNOTIFY;
	tnd.hIcon               = hIcon;

	lstrcpyn(tnd.szTip, strToolTip, sizeof(tnd.szTip));

	Shell_NotifyIcon(dwMsgType, &tnd);

	if (hIcon) 
	  DestroyIcon(hIcon);
}


void poppet_fetch_mail() 
{
  int ret;
    char msg[STRING] = "";

    if (busyDeliver) return;
    else busyDeliver = 1;

    InstallTrayIcon(NIM_MODIFY, TRAYICON_MAIL, "Poppet");
    ret = pop_fetch_mail ( msg, sizeof(msg) );
    if (ret > 0)
      mails += ret;
    if (mails > 0 && FilterError->num==0)
    {
        snprintf( msg, sizeof(msg), "%d Message%s Retrieved", mails, mails>1 ? "s":"");
	InstallTrayIcon(NIM_MODIFY, TRAYICON_RAD, msg);
	if (BeepNew && ret>0)
	    PlaySound("Mailbeep", 0, SND_ALIAS | SND_ASYNC | SND_NOSTOP);
    } 
    else if (ret==0 && FilterError->num==0)
    {
        InstallTrayIcon(NIM_MODIFY, TRAYICON_ENV_01, msg[0] ? msg : "Poppet");
    } 
    else 
    {
        InstallTrayIcon(NIM_MODIFY, TRAYICON_CRS, msg[0] ? msg : 
			FilterError->num == 0 ? "Unkown Error" :
			"Message left in $tmpdir");
    }

    busyDeliver = 0;
}


void poppet_menu(HWND hwnd)
{
    int ret = 0;
    POINT curpos;
    HMENU hSubMenu;
    
    hSubMenu = CreatePopupMenu();
    if (hSubMenu) {
      AppendMenu(hSubMenu, MF_STRING | (busyDeliver ? MF_GRAYED : 0), ID_POPUP_FETCH, "Deliver Now");
#if !defined(NDEBUG) && !defined(CONSOLE)
      AppendMenu(hSubMenu, MF_STRING, ID_POPUP_LOG, "Show Log");
#endif
      /* AppendMenu(hSubMenu, MF_STRING, ID_POPUP_CLEAR, "Clear Icon"); */
      AppendMenu(hSubMenu, MF_STRING, ID_POPUP_ABOUT, "About");
      AppendMenu(hSubMenu, MF_STRING, ID_POPUP_EXIT, "Exit");
      
      GetCursorPos(&curpos);
      SetForegroundWindow(hwnd);
      ret = TrackPopupMenu( hSubMenu,
			    TPM_BOTTOMALIGN | TPM_RIGHTALIGN | TPM_NONOTIFY |
			  TPM_LEFTBUTTON | TPM_RETURNCMD,
			    curpos.x, curpos.y, 0, hwnd, NULL);
      PostMessage(hwnd, WM_NULL, 0, 0);
      
      DestroyMenu(hSubMenu);
    }

    switch (ret) {
      case ID_POPUP_FETCH:
	PostMessage(thehWnd, WM_TIMER, 0, 0);
	break;

      case ID_POPUP_CLEAR:
	break;

      case ID_POPUP_LOG:
#if !defined(NDEBUG) && !defined(CONSOLE)
	SetForegroundWindow(thehWnd);
	DialogBox(GetModuleHandle(NULL),
		  MAKEINTRESOURCE(IDD_STATUS), thehWnd, LogDlgProc);
	PostMessage(thehWnd, WM_NULL, 0, 0);
#endif
	break;

      case ID_POPUP_ABOUT:
	if (busyAbout)
	  break;
	busyAbout = 1;
	DialogBox(GetModuleHandle(NULL),
		  MAKEINTRESOURCE(IDD_ABOUT), thehWnd, AboutDlgProc);
	busyAbout = 0;
	break;

      case ID_POPUP_EXIT:
	PostMessage(thehWnd, WM_DESTROY, 0, 0);
	break;
    }
}

LRESULT CALLBACK
WndProc( HWND hwnd,
	 UINT Message,
	 WPARAM wParam,
	 LPARAM lParam )
{
    NOTIFYICONDATA tnd;
    const UINT TimerSec = 1;


    switch(Message) {
    case WM_CREATE:
	if (quick_exit) {
#if !defined(NDEBUG) && !defined(CONSOLE)
	  DialogBox(GetModuleHandle(NULL),
		    MAKEINTRESOURCE(IDD_STATUS), hwnd, LogDlgProc);
#endif
	  PostQuitMessage(0);
	}
	else {
	  myHwnd = hwnd;
	  InstallTrayIcon(NIM_ADD, TRAYICON_ENV_01, "Poppet");
	  if (PopInterval)
	    SetTimer(hwnd, TimerSec, PopInterval*1000, NULL);
	  PostMessage(hwnd, WM_TIMER, 0, 0);
	}
	break;

    case WM_TRAYNOTIFY:
        if(wParam == IDC_TRAYICON)
	  switch(lParam) {
	    case WM_RBUTTONDOWN:
	      poppet_menu (hwnd);
	      break;

	    case WM_LBUTTONDOWN:
	      {
		SetForegroundWindow(hwnd);

		if (busyHistory)
		  break;
		busyHistory = 1;

		DialogBox(GetModuleHandle(NULL),
			  MAKEINTRESOURCE(IDD_STATUS), hwnd, StatusDlgProc);
		PostMessage(hwnd, WM_NULL, 0, 0);

		busyHistory = 0;
		break;
	      }
	  }
	
	break;
	
    case WM_DESTROY:
        tnd.cbSize   = sizeof(NOTIFYICONDATA);
	tnd.hWnd     = hwnd;
	tnd.uID      = IDC_TRAYICON;
	Shell_NotifyIcon(NIM_DELETE, &tnd);
	KillTimer(hwnd, TimerSec);
	PostQuitMessage(0);
	break;

    case WM_TIMER:
	poppet_fetch_mail();
	break;
	
    default:
        return DefWindowProc(hwnd, Message, wParam, lParam);
    }
    return 0;
}
